"""
Converter plugin management
"""

from __future__ import annotations

import importlib.metadata
from typing import TYPE_CHECKING, Any


if TYPE_CHECKING:
    import ast

    from .converter import Converter, Resolver


class BaseConverterPlugin:
    def init(self, converter: Converter):
        """
        The source file has been parsed, plugins have been loaded, and the conversion
        process is about to start.

        Args:
            converter (Converter): The current converter instance.
        """

    def build_start(self, converter: Converter):
        """
        The new project and app are about to be built.

        Args:
            converter (Converter): The current converter instance.
        """

    def collect_imports(self, converter: Converter):
        """
        Imported symbols have been collected into ``self.imports``.

        Args:
            converter (Converter): The current converter instance.
        """

    def build_project_done(self, converter: Converter):
        """
        The project directory has been created, the settings are about to be modified.

        Args:
            converter (Converter): The current converter instance.
        """

    def build_settings(
        self, converter: Converter, resolver: Resolver, settings_ast: ast.AST
    ) -> tuple[Resolver, ast.AST]:
        """
        Modify the settings AST.

        Args:
            converter (Converter): The current converter instance.
            resolver (Resolver): The current resolver instance for
                ``project/settings.py``.
            settings_ast (ast.AST): The abstract syntax tree for the settings file.
        """
        return resolver, settings_ast

    def build_settings_done(self, converter: Converter):
        """
        The ``project/settings.py`` file has been updated and the assets are about to be
        copied.

        Args:
            converter (Converter): The current converter instance.
        """

    def copy_assets(self, converter: Converter):
        """
        Copy additional assets into the project.

        Args:
            converter (Converter): The current converter instance.
        """

    def build_app_models(
        self, converter: Converter, resolver: Resolver, extra_src: list[str]
    ) -> tuple[Resolver, list[str]]:
        """
        Extend ``app/models.py``.

        Models have been collected on ``converter.models`` (wrapped in ``AppModel``
        instances) and are about to be written into ``app/models.py``.

        Args:
            converter (Converter): The current converter instance.
            resolver (Resolver): The current resolver instance for ``app/models.py``.
            extra_src (list[str]): Lines of Python code to insert at the end.

        Returns:
            resolver (Resolver): The current resolver instance.
            extra_src (list[str]): Lines of Python code to insert at the end.
        """
        return resolver, extra_src

    def build_app_models_done(self, converter: Converter):
        """
        The ``app/models.py`` has been written, the views are about to be built.

        Args:
            converter (Converter): The current converter instance.
        """

    def build_app_views(
        self, converter: Converter, resolver: Resolver, extra_src: list[str]
    ) -> tuple[Resolver, list[str]]:
        """
        Extend ``app/views.py``.

        Views have been collected on ``converter.views`` (wrapped in ``AppView``
        instances) and are about to be written into ``app/views.py``.

        Args:
            converter (Converter): The current converter instance.
            resolver (Resolver): The current resolver instance for ``app/views.py``.
            extra_src (list[str]): Lines of Python code to insert at the end.

        Returns:
            resolver (Resolver): The current resolver instance.
            extra_src (list[str]): Lines of Python code to insert at the end.
        """
        return resolver, extra_src

    def build_app_views_done(self, converter: Converter):
        """
        The ``app/views.py`` has been written, the app urls are about to be built.

        Args:
            converter (Converter): The current converter instance.
        """

    def build_app_api(
        self,
        converter: Converter,
        resolver: Resolver,
        extra_src: list[str],
    ) -> tuple[Resolver, list[str]]:
        """
        Extend ``app/api.py``

        API endpoints have been collected on ``converter.api_views`` (wrapped in
        ``AppAPIView`` instances) and are about to be written into ``app/api.py``

        Args:
            converter (Converter): The current converter instance.
            resolver (Resolver): The current resolver instance for ``app/admin.py``.
            admins (list[str]): Lines of Python code for model admin definitions.
            extra_src (list[str]): Lines of Python code to insert at the end.

        Returns:
            resolver (Resolver): The current resolver instance.
            admins (list[str]): Lines of Python code for model admin definitions.
            extra_src (list[str]): Lines of Python code to insert at the end.
        """
        return resolver, extra_src

    def build_app_api_done(self, converter: Converter):
        """
        The ``app/api.py`` has been written (or not if there were no APIs defined), and
        the unused objects are about to be built.

        Args:
            converter (Converter): The current converter instance.
        """

    def build_app_urls(
        self,
        converter: Converter,
        resolver: Resolver,
        urls: list[str],
        extra_src: list[str],
    ) -> tuple[Resolver, list[str], list[str]]:
        """
        Extend app/urls.py

        Urls have been collected and passed into this hook as ``urls``.

        Args:
            converter (Converter): The current converter instance.
            resolver (Resolver): The current resolver instance for ``app/urls.py``.
            urls (list[str]): Lines of Python to insert into ``urlpatterns = [...]``
            extra_src (list[str]): Lines of Python code to insert at the end.

        Returns:
            resolver (Resolver): The current resolver instance.
            urls (list[str]): Lines of Python to insert into ``urlpatterns = [...]``
            extra_src (list[str]): Lines of Python code to insert at the end.
        """
        return resolver, urls, extra_src

    def build_app_urls_done(self, converter: Converter):
        """
        The ``app/urls.py`` has been written, the project urls are about to be built.

        Args:
            converter (Converter): The current converter instance.
        """

    def build_urls(self, converter: Converter, src: str) -> str:
        """
        Modify ``project/urls.py``

        The ``project/urls.py`` has been loaded into ``src``, and the app's urls (if
        present) and admin URL have been updated. This hook can modify the source
        further, returning the full string to be written out to the file.

        Args:
            converter (Converter): The current converter instance. src (str): The Python
            code for ``project/urls.py``

        Returns:
            src (str): The Python code
        """
        return src

    def build_urls_done(self, converter: Converter):
        """
        The ``project/urls.py`` has been written, the app admin is about to be built.

        Args:
            converter (Converter): The current converter instance.
        """

    def build_app_admin(
        self,
        converter: Converter,
        resolver: Resolver,
        admins: list[str],
        extra_src: list[str],
    ) -> tuple[Resolver, list[str], list[str]]:
        """
        Extend ``app/admin.py``

        ModelAdmin definitions have been built into ``admins``. This hook can modify
        this list, or add some extra lines of source code to add to the end of the file.

        Args:
            converter (Converter): The current converter instance. resolver (Resolver):
            The current resolver instance for ``app/admin.py``. admins (list[str]):
            Lines of Python code for model admin definitions. extra_src (list[str]):
            Lines of Python code to insert at the end.

        Returns:
            resolver (Resolver): The current resolver instance. admins (list[str]):
            Lines of Python code for model admin definitions. extra_src (list[str]):
            Lines of Python code to insert at the end.
        """
        return resolver, admins, extra_src

    def build_app_admin_done(self, converter: Converter):
        """
        The ``app/admin.py`` has been written, the unused code is about to be written
        out.

        Args:
            converter (Converter): The current converter instance.
        """

    def build_app_unused(
        self,
        converter: Converter,
        resolver: Resolver,
        extra_src: list[str],
    ) -> tuple[Resolver, list[str]]:
        """
        Extend ``app/unused.py``

        All active files have been built, any leftover code is about to be written into
        ``app/unused.py``.

        It is better to collect important code before this function.

        Args:
            converter (Converter): The current converter instance.
            resolver (Resolver): The current resolver instance for ``app/unused.py``.
            extra_src (list[str]): Lines of Python code to insert at the end.

        Returns:
            resolver (Resolver): The current resolver instance.
            extra_src (list[str]): Lines of Python code to insert at the end.
        """
        return resolver, extra_src

    def build_end(self, converter: Converter):
        """
        The build process is complete.

        Args:
            converter (Converter): The current converter instance.
        """


class Manager:
    # Entrypoint group
    entrypoint = "nanodjango"

    def __init__(self):
        self.plugins = []

    def load(self):
        """
        Load plugins
        """
        # Load by contrib
        from . import contrib  # noqa

        # Load by entrypoint
        for dist in importlib.metadata.distributions():
            for entry_point in dist.entry_points:
                if entry_point.group != self.entrypoint:
                    continue

                # Import the module - any plugins will auto-register
                entry_point.load()

    def register(self, cls):
        self.plugins.append(cls)

    def __getattribute__(self, name):
        """
        Pass hook calls to the manager through to the plugins
        """

        def call(converter, *args) -> Any:
            is_single_arg = False
            returned = args
            if len(args) == 1:
                is_single_arg = True
                returned = args[0]

            for plugin in self.plugins:
                fn = getattr(plugin, name)
                returned = fn(plugin, converter, *args)
                # TODO: Add type checking to safeguard plugin return values
                if returned is None:
                    pass
                elif is_single_arg:
                    args = (returned,)
                else:
                    args = returned

            return returned

        if name in BaseConverterPlugin.__dict__:
            return call

        return super().__getattribute__(name)


plugins = Manager()


# We need to define the ConverterPlugin after the manager, as it needs to auto-register.
# But the manager needs the ConverterPlugin to be defined before, as it introspects it
# We therefore introspect the BaseConverterPlugin, and add auto-registration afterwards
class ConverterPlugin(BaseConverterPlugin):
    def __init_subclass__(cls) -> None:
        plugins.register(cls)



from types import ModuleType

from django.apps.config import AppConfig
from django.apps.registry import apps as apps_registry

from .. import settings


class NanodjangoAppConfig(AppConfig):
    """
    AppConfig with a hard-coded path
    """

    path = str(settings.DF_FILEPATH)


def prepare_apps(app_name: str, app_module: ModuleType):
    """
    Create and register the app config

    Do this while our script is importing, before apps.populate() is called.

    This tricks the apps registry into thinking our script module has already been
    loaded and doesn't try to import it again. Our models will get registered when the
    module finishes importing.
    """
    app_config = NanodjangoAppConfig(app_name=app_name, app_module=app_module)
    apps_registry.app_configs[app_config.label] = app_config
    app_config.apps = apps_registry
    app_config.models = {}



from __future__ import annotations

import ast


class ReferenceVisitor(ast.NodeVisitor):
    """
    Visitor to traverse the AST of a class or function, looking for references to
    objects outside its scope
    """

    # This class is likely incomplete. Contributions welcome.

    #: Set of global names referenced
    globals_ref: set[str]

    def __init__(self):
        self.locals_stack = [set()]
        self.globals_ref = set()

    def push_scope(self):
        self.locals_stack.append(self.current_scope.copy())

    def pop_scope(self):
        self.locals_stack.pop()

    @property
    def current_scope(self):
        return self.locals_stack[-1]

    @property
    def local_scopes(self):
        return set().union(*self.locals_stack)

    def found_reference(self, ref):
        if ref in __builtins__:
            return
        if ref not in self.local_scopes:
            self.globals_ref.add(ref)

    def visit_FunctionDef(self, node):
        """Function definition, including top level definition if obj is a function"""
        self.current_scope.add(node.name)

        # Find all new local args in fn definition
        args = node.args.args + node.args.kwonlyargs
        if node.args.vararg:
            args.append(node.args.vararg)
        if node.args.kwarg:
            args.append(node.args.kwarg)

        self.push_scope()
        for arg in args:
            self.current_scope.add(arg.arg)

        for decorator in node.decorator_list:
            self.visit(decorator)
        self.generic_visit(node)
        self.pop_scope()

    def visit_ClassDef(self, node):
        """Function definition, including top level definition if obj is a class"""
        self.push_scope()
        for base in node.bases:
            self.visit(base)
        for decorator in node.decorator_list:
            self.visit(decorator)
        for stmt in node.body:
            self.visit(stmt)
        self.pop_scope()

    def visit_Assign(self, node):
        """Direct variable assignments"""
        for target in node.targets:
            if isinstance(target, ast.Name):
                self.current_scope.add(target.id)
        self.generic_visit(node)

    def visit_NamedExpr(self, node):
        """Walrus operator assignment"""
        if isinstance(node.target, ast.Name):
            self.current_scope.add(node.target.id)
        self.visit(node.value)

    def visit_Attribute(self, node):
        """Accessing an attribute of a variable"""
        if isinstance(node.value, ast.Name):
            self.found_reference(node.value.id)
        self.visit(node.value)

    def visit_Name(self, node):
        """Accessing something in scope, eg a variable"""
        if isinstance(node.ctx, ast.Load):
            self.found_reference(node.id)

    def visit_Global(self, node):
        """Bringing in a global reference"""
        for name in node.names:
            self.globals_ref.add(name)

    def visit_ListComp(self, node: ast.ListComp | ast.SetComp | ast.GeneratorExp):
        """List and set comprehensions, and generator expressions"""
        self.push_scope()
        for generator in node.generators:
            self.visit(generator)
        self.visit(node.elt)
        self.pop_scope()

    visit_SetComp = visit_ListComp
    visit_GeneratorExp = visit_ListComp

    def visit_DictComp(self, node):
        """Dictionary comprehension"""
        self.push_scope()
        for generator in node.generators:
            self.visit(generator)
        self.visit(node.key)
        self.visit(node.value)
        self.pop_scope()

    def visit_Import(self, node: ast.Import | ast.ImportFrom):
        for alias in node.names:
            if alias.asname:
                self.current_scope.add(alias.asname)
            else:
                self.current_scope.add(alias.name)

    visit_ImportFrom = visit_Import



from __future__ import annotations

import ast
import inspect
import os
import shutil
import subprocess
from pathlib import Path
from types import ModuleType
from typing import TYPE_CHECKING, cast

from django.db.models import Model

import isort
from black import FileMode, format_str

from ..exceptions import ConversionError
from .objects import AppApiView, AppModel, AppView
from .plugin import plugins
from .utils import (
    collect_references,
    ensure_http_response,
    filter_decorators,
    import_from_path,
    is_api_decorator,
    make_url,
    obj_to_ast,
)


if TYPE_CHECKING:
    from pathlib import Path

    from ..app import Django


class Resolver:
    """
    Resolve names and references within the scope of the Python module we are generating
    """

    imports: set[str]
    local_refs: set[str]
    global_refs: set[str]

    def __init__(self, converter: Converter, module_name: str):
        """
        Args:
            converter (Converter): The current converter instance
            module_name (str): The name of the module we're currently building
        """
        self.converter = converter
        self.module_name = module_name

        self.imports = set()
        self.local_refs = set()
        self.global_refs = set()

    def add(self, name: str, references: set[str]):
        """
        Register an object definition, and the objects it references
        """
        self.add_object(name)
        self.add_references(references)

    def add_object(self, name: str):
        """
        Register an object that is defined in the current module we're building
        """
        self.converter.imports[name] = f"from {self.module_name} import {name}"
        self.local_refs.add(name)

    def add_references(self, references: set[str]):
        """
        Register references to symbols needed in this module - objects which will either
        need to be imported from other modules, or their source added to this module
        """
        for ref in references:
            # Most of the time this will be either an object or an imported symbol,
            # but it could be an object.child, eg admin.site.urls
            ref_base = ref.split(".", 1)[0]
            if ref_base in self.local_refs:
                pass
            elif ref_base in self.converter.imports:
                self.imports.add(self.converter.imports[ref_base])
            else:
                self.global_refs.add(ref_base)

    def gen_src(self):
        """
        Generate the source code required to resolve discovered references, either by
        copying in referenced code, or importing them from another file.
        """
        all_src = []
        global_refs = self.global_refs.copy()
        while self.global_refs:
            # Collect the source for the reference, and grab any references it may have
            global_ref = self.global_refs.pop()
            src, references = self.converter.collect_definition(global_ref)
            all_src.append(src)

            # Record references from this
            self.add_references(references)

            # We may have added references to locals we have seen before - remove them
            self.global_refs = self.global_refs - self.local_refs

            # The global reference is now local
            self.local_refs.add(global_ref)

            # Tell other modules where to find it
            self.converter.imports[global_ref] = (
                f"from {self.module_name} import {global_ref}"
            )

        # Restore global refs
        self.global_refs = global_refs
        return "\n".join(list(self.imports) + all_src)


class Converter:
    #: Reference to the ``Django`` app instance
    app: Django

    #: Root path to build the project (``django-admin startproject ... {root_path}``)
    root_path: Path

    #: Name of the Django project (``django-admin startproject {project_name} ...``)
    project_name: str

    #: The module to convert
    module: ModuleType

    #: The source for the module to convert
    src: str

    #: The abstract syntax tree of the module to convert
    ast: ast.Module

    #: AppModel instances for models that are being moved to models.py
    models: list[AppModel]

    #: AppView instances for views that are being moved to views.py
    views: list[AppView]

    #: Import paths - existing imports in the app, plus converted models and views
    imports: dict[str, str]

    #: Definitions in the module's top level scope which have been converted
    used: set[str]

    def __init__(self, app: Django, path: Path, name: str):
        """
        Prepare state and load plugins
        """
        self.app = app
        self.root_path = path
        self.project_name = name
        self.module = app.app_module
        self.src = inspect.getsource(app.app_module)
        self.ast = ast.parse(self.src)

        self.models = []
        self.views = []
        self.api_views = []
        self.extra_urls = []
        self.imports = {}
        self.used = set()

        plugins.load()
        plugins.init(self)

    @property
    def project_path(self):
        return self.root_path / self.project_name

    @property
    def app_path(self):
        return self.project_path / self.app.app_name

    def write_file(self, filename: str | Path, *content):
        path = Path(filename)
        formatted = isort.code(format_str("\n".join(content), mode=FileMode()))
        path.write_text(formatted)

    def collect_definition(self, obj_name) -> tuple[str, set[str]]:
        """
        Collect a definition of a module top-level scope object

        Args:
            obj_name (str): Name of object to collect

        Returns:
            obj_src (str): Source for object
            references (set[str]): Set of referenced object names
        """
        if obj_name == "ensure_http_response":
            # Inject its dependencies
            self.imports.update(getattr(ensure_http_response, "_dependencies", {}))
            obj_src = inspect.getsource(ensure_http_response)
            obj_ast = ast.parse(obj_src)
            references = collect_references(obj_ast.body[0])
            return obj_src, references

        if obj_name not in self.module.__dict__:
            raise ValueError(f"Reference to unknown symbol {obj_name}")
        obj = self.module.__dict__[obj_name]

        # Try to build src from the object
        try:
            obj_src = inspect.getsource(obj)
        except TypeError:
            # Not a class, method, function, or code object
            pass
        else:
            self.used.add(obj_name)
            obj_ast = ast.parse(obj_src)
            references = collect_references(obj_ast.body[0])
            return obj_src, references

        # Look for an assignment
        for node in self.ast.body:
            if isinstance(node, ast.Assign):
                for target in node.targets:
                    if isinstance(target, ast.Name) and target.id == obj_name:
                        obj_src = ast.unparse(node).strip()
                        self.used.add(obj_name)
                        obj_ast = ast.parse(obj_src)
                        references = collect_references(obj_ast.body[0])
                        return obj_src, references

        # TODO: We could have more exhaustive definition collection
        raise ValueError(f"Reference to undetermined symbol {obj_name}")

    def build(self) -> None:
        """
        Create the project and build the files for the project and app

        Hooks:
            build_start: Called at the start
            build_end: Called at the end
        """
        plugins.build_start(self)

        self.collect_imports()

        self.build_project()
        plugins.build_project_done(self)

        self.build_settings()
        plugins.build_settings_done(self)

        self.copy_assets()

        self.build_app_models()
        plugins.build_app_models_done(self)

        self.build_app_views()
        plugins.build_app_views_done(self)

        self.build_app_api()
        plugins.build_app_api_done(self)

        self.app_has_urls = False
        self.build_app_urls()
        plugins.build_app_urls_done(self)
        self.build_urls()
        plugins.build_urls_done(self)

        self.build_app_admin()
        plugins.build_app_admin_done(self)

        self.build_app_unused()

        plugins.build_end(self)

    def collect_imports(self) -> dict[str, str]:
        """
        Collect a lookup for imported names

        Values are the import strings - we'll use isort to merge modules later
        """
        self.imports = {}
        for node in self.ast.body:
            if isinstance(node, ast.Import):
                for alias in node.names:
                    self.imports[alias.name] = f"import {alias.name}"
            elif isinstance(node, ast.ImportFrom):
                for alias in node.names:
                    self.imports[alias.name] = f"from {node.module} import {alias.name}"

        plugins.collect_imports(self)
        return self.imports

    def build_project(self) -> None:
        """
        Run ``django-admin startproject`` and create the app dir
        """
        # Copy the env, so we're still working within any venv
        env = os.environ.copy()
        env.pop("DJANGO_SETTINGS_MODULE", None)
        try:
            result = subprocess.run(
                ["django-admin", "startproject", self.project_name, self.root_path],
                env=env,
                capture_output=True,
                text=True,
            )
        except Exception as e:
            raise ConversionError(f"Could not run django-admin: {e}")

        if result.returncode != 0:
            raise ConversionError(f"django-admin startproject failed: {result.stderr}")

        # Create app dir
        app_dir = self.app_path
        app_dir.mkdir()
        (app_dir / "__init__.py").touch()

    def build_settings(self) -> None:
        """
        Collect settings from the app definition and update the project settings

        Hooks:
            build_settings: Customise ``project/settings.py`` AST
            build_settings_done: After ``project/settings.py`` has been written
        """
        resolver = Resolver(self, f"{self.project_name}.settings")

        # Collect from app definition and remove nanodjango-specific settings
        app_settings = {}
        for node in self.ast.body:
            # Look for app = Django(..)
            if (
                isinstance(node, ast.Assign)
                and any(
                    [
                        target.id == self.app._instance_name
                        for target in node.targets
                        if isinstance(target, ast.Name)
                    ]
                )
                and isinstance(node.value, ast.Call)
                and isinstance(node.value.func, ast.Name)
                and node.value.func.id == "Django"
            ):
                for keyword in node.value.keywords:
                    name: str = cast(str, keyword.arg)
                    if name.isupper():
                        # Exclude nanodjango-specific settings
                        if name in [
                            "ADMIN_URL",
                            "EXTRA_APPS",
                            "SQLITE_DATABASE",
                            "MIGRATIONS_DIR",
                        ]:
                            continue
                        else:
                            app_settings[name] = keyword.value
                            resolver.add_references(collect_references(keyword.value))

        # Load settings file
        filename = self.project_path / "settings.py"
        settings = import_from_path("nanodjango.convert.tmp_settings", filename)
        settings_src = inspect.getsource(settings)
        settings_ast = ast.parse(settings_src)

        # Replace settings
        for node in settings_ast.body:
            if isinstance(node, ast.Assign):
                for target in node.targets:
                    if not isinstance(target, ast.Name):
                        continue
                    if target.id in app_settings:
                        node.value = app_settings.pop(target.id)

                    # Special case for INSTALLED_APPS
                    if target.id == "INSTALLED_APPS" and isinstance(
                        node.value, ast.List
                    ):
                        # Inject our converted app
                        node.value.elts.append(
                            ast.Constant(
                                value=f"{self.project_name}.{self.app.app_name}"
                            )
                        )

                        # Append any EXTRA_APPS
                        if "EXTRA_APPS" in self.app._settings:
                            for extra_app in self.app._settings["EXTRA_APPS"]:
                                node.value.elts.append(ast.Constant(value=extra_app))

        # Add any extra vars
        for name, value in app_settings.items():
            node = ast.Assign(targets=[ast.Name(id=name, ctx=ast.Store())], value=value)
            settings_ast.body.append(node)

        # Plugin hook
        resolver, settings_ast = plugins.build_settings(self, resolver, settings_ast)

        # Insert any references - usually imports
        ref_src = resolver.gen_src()
        if ref_src:
            imports = set()
            others = set()

            for node in ast.parse(ref_src).body:
                if isinstance(node, (ast.Import, ast.ImportFrom)):
                    imports.add(node)
                else:
                    others.add(node)

            # We're expecting a comment at the top - insert imports just after that
            if (
                len(settings_ast.body) == 0
                or not isinstance(settings_ast.body[0], ast.Expr)
                or not isinstance(settings_ast.body[0].value, ast.Constant)
            ):
                raise ConversionError("Unexpected start to settings.py")
            settings_ast.body[1:1] = imports

            # Insert other definitions after the last import
            index = 0
            for node in settings_ast.body:
                if isinstance(node, (ast.Expr, ast.Import, ast.ImportFrom)):
                    index += 1
                    continue
                break
            settings_ast.body[index:index] = others

        # Save settings
        self.write_file(filename, ast.unparse(settings_ast))

    def copy_assets(self) -> None:
        """
        Copy static, templates, and migrations into app, and db.sqlite3 if it exists

        Hooks:
            copy_assets: After the primary assets have been copied
        """
        script_dir = self.app.app_path.parent

        db_file = script_dir / self.app._settings.get("SQLITE_DATABASE", "db.sqlite3")
        if db_file.exists():
            shutil.copy(db_file, self.root_path / "db.sqlite3")

        # List of (source, dest)
        dir_names = [
            ("static", "static"),
            ("templates", "templates"),
            (self.app._settings.get("MIGRATIONS_DIR", "migrations"), "migrations"),
        ]
        for source_name, dest_name in dir_names:
            src_dir = script_dir / source_name
            if src_dir.exists() and src_dir.is_dir():
                shutil.copytree(
                    src_dir,
                    self.app_path / dest_name,
                )

        plugins.copy_assets(self)

    def build_app_models(self) -> None:
        """
        Build ``app/models.py`` and collect models in ``self.models`` (a list of
        ``AppModel instances).

        Hooks:
            build_app_models: Modify ``self.models`` or add content to ``app/models.py``
            build_app_models_done: After ``app/models.py`` has been written
        """
        self.models = []
        resolver = Resolver(self, ".models")

        for name, obj in self.module.__dict__.items():
            if inspect.isclass(obj) and issubclass(obj, Model):
                app_model = AppModel(self, name, obj)
                self.models.append(app_model)
                resolver.add(name, app_model.references)

        resolver, extra_src = plugins.build_app_models(self, resolver, [])

        if not self.models and not extra_src:
            return

        self.write_file(
            self.app_path / "models.py",
            resolver.gen_src(),
            "\n".join([app_model.src for app_model in self.models]),
            "\n".join(extra_src),
        )

    def build_app_views(self) -> None:
        """
        Build ``app/views.py`` and collect views in ``self.views`` (a list of
        ``AppView`` instances).

        Hooks:
            build_app_views: Modify ``self.views`` or add content to ``app/views.py``
            build_app_views_done: After ``app/views.py`` has been written.
        """
        self.views = []
        resolver = Resolver(self, ".views")

        for pattern, (view, url_config) in self.app._routes.items():
            if view is None:
                # An include
                continue
            app_view = AppView(self, view, pattern, url_config)
            self.views.append(app_view)
            resolver.add(view.__name__, app_view.references)

        resolver, extra_src = plugins.build_app_views(self, resolver, [])

        if not self.views and not extra_src:
            return

        self.write_file(
            self.app_path / "views.py",
            resolver.gen_src(),
            "\n".join([app_view.src for app_view in self.views]),
            "\n".join(extra_src),
        )

    def build_app_api(self) -> None:
        # TODO #7
        self.api_views = []
        resolver = Resolver(self, ".api")

        # API definitions will be rewritten from @app.api to @api, and we'll hard-code
        # the ``api`` definition when writing this module at the end of this method,
        # so tell the resolver we already know about ``api``
        resolver.add_object("api")

        for name, obj in self.module.__dict__.items():
            # Look at locally-defined functions
            if (
                inspect.isfunction(obj)
                and inspect.getsourcefile(obj) == self.module.__file__
            ):
                obj_src = inspect.getsource(obj)
                obj_ast = cast(ast.FunctionDef, obj_to_ast(obj_src))
                api_decorators, _ = filter_decorators(
                    obj_ast,
                    is_api_decorator,
                    self.app._instance_name,
                )
                if not api_decorators:
                    continue

                api_view = AppApiView(
                    self,
                    name=name,
                    obj=obj,
                    obj_src=obj_src,
                    obj_ast=obj_ast,
                )
                self.api_views.append(api_view)
                resolver.add(name, api_view.references)

        resolver, extra_src = plugins.build_app_api(self, resolver, [])
        if not self.api_views and not extra_src:
            return

        self.app_has_urls = True
        self.write_file(
            self.app_path / "api.py",
            "from ninja import NinjaAPI",
            resolver.gen_src(),
            "api = NinjaAPI()",
            "\n".join([api_view.src for api_view in self.api_views]),
            "\n".join(extra_src),
        )

    def build_app_urls(self) -> None:
        """
        Build ``app/urls.py``

        Hooks:
            build_app_views: Modify ``self.views`` or add content to ``app/views.py``
            build_app_views_done: After ``app/views.py`` has been written.
        """
        imports = set()
        urls = []
        resolver = Resolver(self, ".urls")

        # Add API url - this is a special case which never makes it into app._routes
        if self.api_views:
            urls.append(make_url(self.app.settings.API_URL, "api.urls"))
            resolver.add_references(["api"])

        # Add view urls
        app_views = self.views.copy()
        for pattern, (view, url_config) in self.app._routes.items():
            if app_views and app_views[0].pattern == pattern:
                # path(pattern, view)
                app_view = app_views.pop(0)
                urls.append(app_view.make_url())
            else:
                # path(pattern, include)
                # Extract the ``include`` reference
                route_ast = url_config["source"]
                if (
                    isinstance(route_ast, ast.Expr)
                    and (call := getattr(route_ast, "value"))
                    and isinstance(call, ast.Call)
                    and call.keywords
                    and (
                        include_ast := cast(
                            ast.Attribute,
                            next(kw for kw in call.keywords if kw.arg == "include"),
                        )
                    )
                ):
                    # Collect references, in case this is an include(..) call etc.
                    include_src = ast.unparse(include_ast.value)
                    references = collect_references(include_ast.value)
                    urls.append(make_url(pattern, include_src, **url_config))
                    resolver.add_references(references)
                else:
                    raise ConversionError(
                        f"Could not understand route {url_config['source']}"
                    )
            if url_config["re"]:
                imports.add("re_path")
            else:
                imports.add("path")

        resolver, urls, extra_src = plugins.build_app_urls(self, resolver, urls, [])

        if not urls and not extra_src:
            return

        # Register that we found URLs so build_urls() will link to app.urls.urlpatterns
        self.app_has_urls = True

        views_import = "from . import views"
        if not self.views:
            # We've got a urls.py which doesn't have any views - eg a plugin has added
            # extra_views or given us extra URLs
            views_import = ""

        self.write_file(
            self.app_path / "urls.py",
            f"from django.urls import {', '.join(imports)}",
            views_import,
            resolver.gen_src(),
            "urlpatterns = [",
            "\n".join(self.extra_urls),
            "\n".join(urls),
            "]",
            "\n".join(extra_src),
        )

    def build_urls(self) -> None:
        """
        Update ``project/urls.py``

        Hooks:
            build_urls: Modify source for ``project/urls.py``
            build_urls_done: After ``project/urls.py`` has been written.
        """
        filename = self.project_path / "urls.py"
        src = filename.read_text()

        # Add path to app
        pattern = "urlpatterns = ["
        if pattern not in src:
            raise ConversionError("Expected to find urlpatterns in urls.py")

        if self.app_has_urls:
            src = src.replace(
                "from django.urls import path",
                "from django.urls import include, path",
            ).replace(
                pattern,
                f'{pattern}\n    path("", include("{self.project_name}.{self.app.app_name}.urls")),',
            )

        if "ADMIN_URL" in self.app._settings:
            pattern = '"admin/"'
            if pattern not in src:
                raise ConversionError("Expected to find admin path in urls.py")
            src = src.replace(pattern, f'"{self.app._settings["ADMIN_URL"]}"')

        src = plugins.build_urls(self, src)
        self.write_file(filename, src)

    def build_app_admin(self) -> None:
        """
        Write discovered model admin registrations in ``app/admin.py``

        Hooks:
            build_app_admin: Modify or add content to ``app/admin.py``
            build_app_admin_done: After ``app/admin.py`` has been written.
        """
        model_names: list[str] = []
        admins: list[str] = []
        resolver = Resolver(self, ".admin")

        # Collect @app.admin decorated models
        for app_model in self.models:
            if not app_model.admin_decorator:
                continue
            model_names.append(app_model.name)
            admins.append(app_model.make_model_admin())
            resolver.add_references(set([app_model.name]))

        # TODO: We could collect ModelAdmin. For now lets let it fall into unused

        resolver, admins, extra_src = plugins.build_app_admin(
            self, resolver, admins, []
        )

        if not model_names:
            return

        self.write_file(
            self.app_path / "admin.py",
            "from django.contrib import admin",
            resolver.gen_src(),
            "\n".join(admins),
            "\n".join(extra_src),
        )

    def build_app_unused(self) -> None:
        """
        Write unused code into``app/unused.py``

        Hooks:
            build_app_unused: Modify or add content to ``app/unused.py``
        """
        resolver = Resolver(self, ".unused")

        # We're not going to worry about unused imports
        self.used.update(self.imports.keys())

        # Also going to ignore the app
        self.used.add(self.app._instance_name)

        unused = set(self.module.__dict__.keys()) - self.used
        all_src = []
        for obj_name in unused:
            if obj_name.startswith("_"):
                continue

            # Collect the source and objects it references
            obj_src, references = self.collect_definition(obj_name)
            all_src.append(obj_src)
            resolver.add(obj_name, references)

        resolver, extra_src = plugins.build_app_unused(self, resolver, [])

        if not all_src or extra_src:
            return

        self.write_file(
            self.app_path / "unused.py",
            "# Definitions that were not used by the nanodjango converter",
            "# These will need to be merged into the rest of the app manually",
            resolver.gen_src(),
            "\n".join(all_src),
            "\n".join(extra_src),
        )

        print(f"Unused code detected, see {self.app.app_name}/unused.py")



from __future__ import annotations

import ast
import inspect
import os
import sys
from pathlib import Path
from types import ModuleType
from typing import TYPE_CHECKING, Any, Callable

from django import setup
from django import urls as django_urls
from django.contrib import admin
from django.contrib.auth import get_user_model
from django.db.models import Model
from django.views import View

from . import app_meta
from .exceptions import ConfigurationError, UsageError
from .urls import urlpatterns
from .views import string_view

if TYPE_CHECKING:
    from pathlib import Path

    from ninja import NinjaAPI


def exec_manage(*args):
    from django.core.management import execute_from_command_line

    args = ["nanodjango"] + list(args)
    execute_from_command_line(args)


class Django:
    """
    The main Django app
    """

    # Class variable to ensure there can be only one
    _instantiated = False

    #: Name of the app script
    app_name: str

    #: Reference to the app script's module
    app_module: ModuleType

    #: Path of app script
    app_path: Path

    #: Whether this app has defined an @app.admin
    has_admin: bool = False

    #: Variable name for this current app
    _instance_name: str

    # Settings cache to aid ``convert``
    _settings: dict[str, Any]

    # URL cache to aid ``convert``
    # {pattern:
    _routes: dict[str, tuple[Callable | None, dict[str, Any]]]

    # NinjaAPI instance for @app.api
    _api: NinjaAPI | None = None

    def __new__(cls, *args, **kwargs):
        # Enforce only one Django() per script, otherwise everything will get confused
        if cls._instantiated:
            raise ConfigurationError("An app can only have one Django() instance")
        cls._instantiated = True

        instance = super().__new__(cls)

        # Set app meta
        app_name = inspect.stack()[1].frame.f_globals["__name__"]
        app_meta._app_module = sys.modules[app_name]
        return instance

    def __init__(self, **_settings):
        """
        Initialise a new Django app, optionally with settings

        Usage::

            app = Django()
            app = Django(SECRET_KEY="some-secret", ALLOWED_HOSTS=["my.example.com"])
        """
        self.has_admin = False
        self._settings = {}
        self._routes = {}
        self._config(_settings)

    def _config(self, _settings):
        """
        Configure settings and patch Django ready for model definitions
        """
        self._settings = app_meta._app_conf = _settings

        # Settings
        os.environ.setdefault("DJANGO_SETTINGS_MODULE", "nanodjango.settings")
        from django.conf import settings

        self.settings = settings

        for key, value in _settings.items():
            setattr(settings, key, value)

        self.app_name = settings.DF_APP_NAME
        self.app_module = app_meta.get_app_module()
        self.app_path = Path(inspect.getfile(self.app_module))

        # Import and apply glue after django.conf has its settings
        from .django_glue.apps import prepare_apps
        from .django_glue.db import patch_db

        patch_db(self.app_name)
        prepare_apps(self.app_name, self.app_module)

        # Ready for Django's standard setup
        setup()

    def route(self, pattern: str, *, re=False, include=None, name=None):
        """
        Decorator to add a view to the urls

        The pattern should use the Django path syntax - see
        https://docs.djangoproject.com/en/dev/ref/urls/#path

        Usage::

            # No parameters
            @app.route("/")
            def view(request):
                return "Hello"

            # path() parameters
            @app.route("/<int:pk>/")
            def view(request, pk):
                return f"Hello {pk}"

            # re_path() parameters
            @app.route("/(?P<slug>[a-z]/", re=True)
            def view(request, char):
                return f"Hello {char}"

            # Include another urlconf
            # Note this is called as a function, not a decorator
            # If this is a list not an include()
            app.route("/api/", include=api.urls)

        All paths are relative to the root URL, leading slashes will be ignored.
        """
        # We want to support Flask-like patterns which have leading / in its patterns.
        # Django does not use these, so strip them out.
        pattern = pattern.removeprefix("/")

        # Find if it's a path() or re_path()
        path_fn = django_urls.re_path if re else django_urls.path

        if include is not None:
            # Being called directly with an include
            urlpatterns.append(path_fn(pattern, include))

            # If we're converting, we're going to need the source AST node
            # Get the full source code, then fine the expression by line number
            caller_frame = inspect.currentframe().f_back
            caller_lineno = caller_frame.f_lineno
            caller_source_lines, _ = inspect.findsource(caller_frame)
            caller_source = "".join(caller_source_lines)
            caller_ast = ast.parse(caller_source)

            # Find the node corresponding to the line that called the function
            source = None
            for node in ast.walk(caller_ast):
                if isinstance(node, ast.Expr) and node.lineno == caller_lineno:
                    source = node
                    break

            self._routes[pattern] = (
                None,
                {"re": re, "include": True, "source": source},
            )

            # Make sure this isn't being used as a decorator, that wouldn't make sense
            def invalid(fn):
                raise UsageError(
                    "app.route(path, include=urlconf) cannot be used as a decorator"
                )

            return invalid

        def wrapped(fn):
            # Store route for convert lookup
            self._routes[pattern] = (
                fn,
                {"re": re, "include": False, "name": name or fn.__name__.lower()},
            )

            # Prepare CBVs
            if inspect.isclass(fn) and issubclass(fn, View):
                fn = fn.as_view()

            # Register URL
            urlpatterns.append(
                path_fn(pattern, string_view(fn), name=name or fn.__name__.lower())
            )
            return fn

        return wrapped

    def admin(self, model: type[Model] | None = None, **options):
        """
        Decorator to add a model to the admin site

        The admin site must be added using ``settings.ADMIN_URL``.
        """
        self.has_admin = True

        def wrap(model: type[Model]):
            admin.site.register(model, **options)
            return model

        if model is None:
            # Called with arguments, @admin(attr=val)
            return wrap

        # Called without arguments, @admin - call wrapped immediately
        return wrap(model)

    @property
    def ninja(self):
        """
        Make django-ninja available without needing to import - work around
        https://github.com/vitalik/django-ninja/issues/1169
        """
        try:
            import ninja
        except ImportError as e:
            raise ImportError(
                "Could not find django-ninja - try: pip install django-ninja"
            ) from e
        return ninja

    @property
    def api(self):
        """
        Ninja integration
        """
        if not self._api:
            api = self.ninja.NinjaAPI()
            self._api = api

        return self._api

    def _prepare(self, with_static=False):
        """
        Perform any final setup for this project after it has been imported:

        * detect if it has been run directly; if so, register it as an app
        * register the admin site

        If with_static is True, serve STATIC_URL and MEDIA_URL using
        django.conf.urls.static.static
        """

        # Check if this is being called from click commands or directly
        if self.app_name not in sys.modules:
            # Hasn't been run through the ``nanodjango`` command
            if (
                "__main__" not in sys.modules
                or getattr(sys.modules["__main__"], self._instance_name) != self
            ):
                # Doesn't look like it was run directly either
                raise UsageError("App module not initialised")

            # Run directly, so register app module so Django won't try to load it again
            sys.modules[self.app_name] = sys.modules["__main__"]

        # Register the admin site
        admin_url = self.settings.ADMIN_URL
        if admin_url or self.has_admin:
            if admin_url is None:
                admin_url = "admin/"
            if not isinstance(admin_url, str) or not admin_url.endswith("/"):
                raise ConfigurationError(
                    "settings.ADMIN_URL must be a string path ending in /"
                )
            urlpatterns.append(
                django_urls.path(admin_url.removeprefix("/"), admin.site.urls)
            )

        # Register the API, if defined
        if self._api:
            self.route(self.settings.API_URL, include=self._api.urls)

        # Register static and media
        if with_static:
            from django.conf.urls.static import static

            if self.settings.STATIC_ROOT and Path(self.settings.STATIC_ROOT).exists():
                urlpatterns.extend(
                    static(
                        self.settings.STATIC_URL,
                        document_root=self.settings.STATIC_ROOT,
                    )
                )
            if self.settings.MEDIA_ROOT and Path(self.settings.MEDIA_ROOT).exists():
                urlpatterns.extend(
                    static(
                        self.settings.MEDIA_URL, document_root=self.settings.MEDIA_ROOT
                    )
                )

    def run(self, args: list[str] | tuple[str] | None = None):
        """
        Run a Django management command, passing all arguments

        Defaults to:
            runserver 0:8000
        """
        # Be helpful and check sys.argv for args. This will almost certainly be because
        # it's running directly.
        if args is None:
            args = sys.argv[1:]

        self._prepare(with_static=True)
        if args:
            exec_manage(*args)
        else:
            exec_manage("runserver", "0:8000")

    def start(self, host: str | None = None):
        """
        Perform app setup commands and run the server
        """
        # Be helpful and check sys.argv for the host
        if host is None:
            print(sys.argv)
            if len(sys.argv) > 2:
                raise UsageError("Usage: start [HOST]")
            elif len(sys.argv) == 2:
                host = sys.argv[1]
        if not host:
            host = "0:8000"

        self._prepare(with_static=True)
        exec_manage("makemigrations", self.app_name)
        exec_manage("migrate")
        User = get_user_model()
        if User.objects.count() == 0:
            exec_manage("createsuperuser")
        exec_manage("runserver", host)

    def convert(self, path: Path, name: str):
        from .convert import Converter

        if path.exists():
            raise UsageError("Upgrade path is not empty - path cannot exist")
        path.mkdir()

        converter = Converter(app=self, path=path, name=name)
        converter.build()

    def __call__(self, *args, **kwargs):
        from django.core.wsgi import get_wsgi_application

        if "DEBUG" not in self._settings:
            from django.conf import settings

            settings.DEBUG = False

        application = get_wsgi_application()
        return application(*args, **kwargs)



from __future__ import annotations

import ast
import inspect
from typing import TYPE_CHECKING, Any, Callable, cast

from django.http import HttpResponse

from .utils import (
    applied_ensure_http_response,
    collect_references,
    filter_decorators,
    is_admin_decorator,
    is_api_decorator,
    is_view_decorator,
    make_url,
    obj_to_ast,
    parse_admin_decorator,
)


if TYPE_CHECKING:
    from .converter import Converter


class ConverterObject:
    def __init__(self, converter: Converter, name: str, obj):
        self.converter = converter
        self.name = name
        self.obj = obj
        self.src = self.src_orig = inspect.getsource(obj)
        self.ast: ast.AST = obj_to_ast(self.src)
        self.references: set[str] = set()

    def remove_decorators(self, filter_fn: Callable) -> list[ast.AST]:
        """
        Remove certain decorators from the object, update self.src and self.ast, and
        return a list of removed decorators.
        """
        self.ast = cast(ast.FunctionDef | ast.ClassDef, self.ast)
        filtered_decorators, other_decorators = filter_decorators(
            self.ast,
            filter_fn,
            app_name=self.converter.app._instance_name,
        )

        self.ast.decorator_list = other_decorators
        self.src = ast.unparse(self.ast)

        return filtered_decorators

    def collect_references(self):
        self.references = collect_references(self.ast)


class AppView(ConverterObject):
    pattern: str
    url_config: dict[str, Any]

    def __init__(
        self,
        converter: Converter,
        obj: Callable,
        pattern: str,
        url_config: dict[str, Any],
    ):
        super().__init__(converter, obj.__name__, obj)
        self.pattern = pattern
        self.url_config = url_config

        # Clear all route decorators
        self.remove_decorators(is_view_decorator)

        self.fix_return_value()
        self.collect_references()

    def fix_return_value(self):
        """
        nanodjango allows string return values; this method tries to ensure the views
        that we generate for Django return an HttpResponse
        """
        # We don't decorate CBVs
        if inspect.isclass(self.obj):
            return

        # Find annotation
        if not callable(self.obj):
            raise ValueError("Unrecognised object registered as a view route")

        annotation = inspect.signature(self.obj).return_annotation
        if issubclass(annotation, HttpResponse):
            # We're already returning an HttpResponse, no changes to make
            return

        # We could be returning a string - add the decorator
        self.ast = cast(ast.FunctionDef, self.ast)
        self.ast.decorator_list.append(applied_ensure_http_response)
        self.src = ast.unparse(self.ast)

    def make_url(self) -> str:
        view_fn = f"views.{self.name}"
        if inspect.isclass(self.obj):
            view_fn = f"{view_fn}.as_view()"

        return make_url(self.pattern, view_fn, **self.url_config)


class AppModel(ConverterObject):
    admin_decorator: ast.AST | None

    def __init__(self, converter: Converter, name, obj):
        super().__init__(converter, name, obj)

        # Find any admin decorator
        admin_decorators = self.remove_decorators(is_admin_decorator)
        if len(admin_decorators) > 1:
            raise ValueError(
                f"Found more than one admin decorator on model {self.name}"
            )
        self.admin_decorator = admin_decorators[0] if admin_decorators else None

        self.collect_references()

    def __repr__(self) -> str:
        return f"<AppModel: {self.name}>"

    def make_model_admin(self) -> str:
        if not self.admin_decorator or not (
            options := parse_admin_decorator(
                self.admin_decorator,
                app_name=self.converter.app._instance_name,
            )
        ):
            return f"admin.site.register({self.name})"

        lines = [
            f"@admin.register({self.name})",
            f"class {self.name}Admin(admin.ModelAdmin):",
        ]
        lines.extend([f"    {key} = {value}" for key, value in options.items()])

        return "\n".join(lines)


class AppApiView(ConverterObject):
    def __init__(
        self,
        converter: Converter,
        name: str,
        obj: Any,
        obj_src: str,
        obj_ast: ast.AST,
    ):
        self.converter = converter
        self.name = name
        self.obj = obj
        self.src = self.src_orig = obj_src
        self.ast: ast.AST = obj_ast
        self.references: set[str] = set()

        # Process
        self.fix_return_value()
        self.collect_references()

    def fix_return_value(self):
        """
        Rewrite @app.api.method to @api.method

        AST structure is:

            Call(
                func=Attribute(
                    value=Attribute(
                        value=Name(id='app'),
                        attr='api',
                    ),
                    attr='get',
                args=[Constant(value='/add')],
                keywords=[]
            )
        """
        for decorator in self.ast.decorator_list:
            if is_api_decorator(decorator, app_name=self.converter.app._instance_name):
                decorator.func.value = decorator.func.value.value
                decorator.func.value.id = "api"

        self.src = ast.unparse(self.ast)



import shutil
import sys
from importlib import import_module
from importlib.machinery import SourceFileLoader
from pathlib import Path

import click

from .app import Django


def load_module(module_name: str, path: str | Path):
    module = SourceFileLoader(module_name, str(path)).load_module()
    return module


def load_app(ctx: click.Context, param: str, value: str) -> Django:
    path = Path(value).absolute()

    # Look for the app name
    script_name = path.name
    app_name = None
    if ":" in str(script_name):
        script_name, app_name = script_name.split(":", 1)
        path = path.parent / script_name

    # Find the app module
    if "." in script_name:
        if path.exists():
            sys.path.append(str(path.parent))
            module = load_module(path.stem, path)
        else:
            raise click.UsageError(f"App {value} is not a file or module")
    else:
        try:
            module = import_module(script_name)
        except ModuleNotFoundError:
            raise click.UsageError(f"App {value} is not a file or module")

    # Find the Django instance to use - first try the app name provided
    if app_name and (app := getattr(module, app_name, None)):
        if isinstance(app, Django):
            return app
        else:
            raise click.UsageError(f"App {app_name} is not a Django instance")

    # None provided, find it
    app = None
    for var, val in module.__dict__.items():
        if isinstance(val, Django):
            app_name = var
            app = val
            break

    if app_name is None or app is None:
        raise click.UsageError(f"App {value} has no Django instances")

    app._instance_name = app_name
    return app


@click.group()
def cli():
    pass


@cli.command()
@click.argument("app", type=str, required=True, callback=load_app)
@click.argument("args", type=str, required=False, nargs=-1)
def run(app: Django, args: tuple[str]):
    """
    Run a management command.

    If no command is specified, it will run runserver 0:8000
    """
    app.run(args)


@cli.command()
@click.argument("app", type=str, required=True, callback=load_app)
@click.argument("host", type=str, required=False, default="")
def start(app: Django, host: str):
    """
    Start the app on the specified IP and port

    This will perform a series of setup commands:

        makemigrations <app>
        migrate
        createsuperuser
        runserver HOST
    """
    app.start(host)


@cli.command()
@click.argument("app", type=str, required=True, callback=load_app)
@click.argument("path", type=click.Path(), required=True)
@click.option("--name", "-n", default="project", help="The project name")
@click.option(
    "--delete",
    "can_delete",
    is_flag=True,
    default=False,
    help="If the target path is not empty, delete it before proceeding",
)
@click.option("--plugin", "-p", multiple=True, help="Converter plugins to load")
def convert(
    app: Django, path: click.Path, name: str, can_delete: bool, plugin: list[str]
):
    # Load plugins
    for index, plugin_path in enumerate(plugin):
        plugin_name = Path(plugin_path).stem
        load_module(
            f"nanodjango.convert.contrib.runtime_{index}_{plugin_name}", plugin_path
        )

    # Clear out target path
    target_path: Path = Path(str(path)).resolve()
    if can_delete and target_path.exists():
        shutil.rmtree(str(target_path))

    app.convert(target_path, name)


def invoke():
    cli(obj={})



from __future__ import annotations

import ast
import importlib.util
import inspect
from functools import wraps
from pathlib import Path
from types import ModuleType
from typing import Any, Callable, cast

from django.http import HttpResponse

from .reference import ReferenceVisitor


def pp_ast(obj_ast: ast.AST):
    """
    Development aid to pretty-print an AST object
    """
    print(ast.dump(obj_ast, indent=2))


def import_from_path(module_name: str, file_path: Path) -> ModuleType:
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    if not spec or spec.loader is None:
        raise ValueError(f"spec_from_file_location failed for {file_path}")
    module = importlib.util.module_from_spec(spec)
    return module


def obj_to_ast(obj_src: str) -> ast.AST:
    """
    Convert a single object into its AST

    ast.parse returns an ast.Module with a body of 1; this confirms a single object was
    parsed, extracts and returns it
    """
    obj_ast_module = ast.parse(obj_src)
    if not isinstance(obj_ast_module, ast.Module) or len(obj_ast_module.body) != 1:
        raise ValueError(
            f"Object source does not seem to be a single object: {obj_src}"
        )
    obj_ast = obj_ast_module.body[0]
    return obj_ast


def collect_references(node: ast.AST) -> set[str]:
    visitor = ReferenceVisitor()
    visitor.visit(node)
    return visitor.globals_ref


def get_decorators(obj_ast: ast.AST) -> list[ast.AST]:
    all_decorators = getattr(obj_ast, "decorator_list", [])
    return all_decorators


def filter_decorators(
    obj_ast: ast.FunctionDef | ast.ClassDef, filter_fn: Callable, app_name: str
) -> tuple[list[ast.AST], list[ast.AST]]:
    matches = []
    others = []
    for decorator in get_decorators(obj_ast):
        if filter_fn(decorator, app_name=app_name):
            matches.append(decorator)
        else:
            others.append(decorator)
    return matches, others


def parse_admin_decorator(obj_ast: ast.AST, app_name: str) -> dict[str, Any]:
    keywords: dict[str, Any] = {}
    if isinstance(obj_ast, ast.Call):
        keywords = {}

        for keyword in obj_ast.keywords:
            if not isinstance(keyword.arg, str):
                raise ValueError(
                    f"Unrecognised @{app_name}.admin argument: {ast.unparse(keyword)}"
                )
            keywords[keyword.arg] = ast.unparse(keyword.value)
        obj_ast = obj_ast.func

    if (
        isinstance(obj_ast, ast.Attribute)
        and isinstance(obj_ast.value, ast.Name)
        and obj_ast.value.id == app_name
        and obj_ast.attr == "admin"
    ):
        return keywords

    raise ValueError(
        f"Unrecognised @{app_name}.admin definition: {ast.unparse(obj_ast)}"
    )


def mk_app_decorator_filter(attr_name: str) -> Callable:
    def filter_fn(obj_ast: ast.AST, app_name: str) -> bool:
        # Match by string - avoids having to deal with AST complexities caused by
        # * calls, eg @app.route(...)
        # * nested decorators, eg @app.api.get(...)
        src = ast.unparse(obj_ast).strip()
        seek = f"{app_name}.{attr_name}"
        if src.startswith(seek) and (len(src) == len(seek) or src[len(seek)] in ".("):
            return True
        return False

    return filter_fn


is_admin_decorator = mk_app_decorator_filter("admin")
is_view_decorator = mk_app_decorator_filter("route")
is_api_decorator = mk_app_decorator_filter("api")


def ensure_http_response(view_fn):
    """
    If a view returns a plain string value, convert it into an HttpResponse
    """

    @wraps(view_fn)
    def wrapped(*args, **kwargs):
        response = view_fn(*args, **kwargs)
        if isinstance(response, HttpResponse):
            return response
        return HttpResponse(response)

    return wrapped


# Dependencies for ReferenceVisitor to find
setattr(
    ensure_http_response,
    "_dependencies",
    {
        "wraps": "from functools import wraps",
        "HttpResponse": "from django.http import HttpResponse",
    },
)


@ensure_http_response
def decorated():
    pass


applied_ensure_http_response = cast(
    ast.FunctionDef, obj_to_ast(inspect.getsource(decorated))
).decorator_list[0]


# Generate a line for an url_patterns
def make_url(pattern, view_fn, re=False, include=None, name=None, **url_config):
    # TODO: We should probably escape self.pattern, but it's an extreme edge case
    # that doesn't seem worth the effort at the moment. Contributions welcome.
    if re:
        path_fn = "re_path"
        r = "r"
    else:
        path_fn = "path"
        r = ""

    if name is not None:
        name_arg = f', name="{name}"'
    else:
        name_arg = ""

    return f'    {path_fn}({r}"{pattern}", {view_fn}{name_arg}),'



"""
Django settings for nanodjango project
"""

from os import getenv
from pathlib import Path
from types import ModuleType

from nanodjango.app_meta import get_app_conf, get_app_module


app_conf = get_app_conf()

# Find paths
DF_APP_MODULE: ModuleType = get_app_module()
if not DF_APP_MODULE or not DF_APP_MODULE.__file__:
    raise ValueError("Invalid app module - incorrect initialisation")
DF_FILEPATH: Path = Path(DF_APP_MODULE.__file__).absolute()
DF_APP_NAME: str = DF_FILEPATH.stem
BASE_DIR: Path = DF_FILEPATH.parent

MIGRATION_MODULES = {DF_APP_NAME: app_conf.get("MIGRATIONS_DIR", "migrations")}

# Standard Django settings
SECRET_KEY = getenv("DJANGO_SECRET_KEY", "not-a-secret")
DEBUG = True
ALLOWED_HOSTS = ["*"]
SITE_ID = 1

# Application definition
INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
] + app_conf.get("EXTRA_APPS", [])

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ROOT_URLCONF = "nanodjango.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [str(BASE_DIR / "templates")],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ]
        },
    }
]

WSGI_APPLICATION = "nanodjango.wsgi.application"

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": BASE_DIR / app_conf.get("SQLITE_DATABASE", "db.sqlite3"),
    }
}
DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator"
    },
    {"NAME": "django.contrib.auth.password_validation.MinimumLengthValidator"},
    {"NAME": "django.contrib.auth.password_validation.CommonPasswordValidator"},
    {"NAME": "django.contrib.auth.password_validation.NumericPasswordValidator"},
]


# Internationalization
LANGUAGE_CODE = "en-us"
TIME_ZONE = "UTC"
USE_I18N = True
USE_L10N = True
USE_TZ = True

# Static files
STATIC_URL = "/static/"
STATICFILES_DIRS = [
    BASE_DIR / "static",
]

# Media files
MEDIA_URL = "/media/"
MEDIA_ROOT = BASE_DIR / "media"


# nanodjango specific config
ADMIN_URL = None
API_URL = "api/"



# App metadata
#
# These values are set by Django.__new__ and Django.__init__ before
# nanodjango.settings is imported, so we can configure settings based on these values
#
# Always access the values through their get_*() methods
from types import ModuleType
from typing import Any


#: Reference to app module
#:
#: Available after the app is imported, before app.run() is called
#:
#: Always access through get_app_module()
_app_module: ModuleType | None = None

#: App configuration
#:
#: Used to configure settings
_app_conf: dict[str, Any] = {}


def get_app_module() -> ModuleType:
    global _app_module
    if _app_module is None:
        raise ValueError("App module can only be accessed after it is imported")
    return _app_module


def get_app_conf() -> dict[str, Any]:
    return _app_conf



"""
Support direct use of django-ninja

This plugin extends nanodjango's built-in support for django-ninja by detecting direct
import and initialisation of NinjaAPI in your code, and moving it into api.py.
"""

from __future__ import annotations

import ast

from ..converter import Converter, Resolver
from ..plugin import ConverterPlugin
from ..utils import collect_references, get_decorators


class NinjaConverter(ConverterPlugin):
    def build_app_api(
        self, converter: Converter, resolver: Resolver, extra_src: list[str]
    ) -> tuple[Resolver, list[str]]:
        """
        Find uses of NinjaAPI instances and move them into api.py
        """
        api_objs = set()

        for obj_ast in converter.ast.body:
            is_ninja = False

            if (
                isinstance(obj_ast, ast.Assign)
                and isinstance(obj_ast.value, ast.Call)
                and isinstance(obj_ast.value.func, ast.Name)
                and obj_ast.value.func.id == "NinjaAPI"
            ):
                for target in obj_ast.targets:
                    if isinstance(target, ast.Name):
                        name = target.id
                        api_objs.add(name)
                        resolver.add_object(name)

                is_ninja = True

            elif isinstance(obj_ast, ast.FunctionDef):
                decorators = get_decorators(obj_ast)
                for decorator in decorators:
                    if isinstance(decorator, ast.Call):
                        decorator = decorator.func

                    if (
                        isinstance(decorator, ast.Attribute)
                        and isinstance(decorator.value, ast.Name)
                        and decorator.value.id in api_objs
                    ):
                        resolver.add_object(obj_ast.name)
                        is_ninja = True

            if is_ninja:
                src = ast.unparse(obj_ast)
                references = collect_references(obj_ast)
                resolver.add_references(references)
                extra_src.append(src)

        return resolver, extra_src



from .utils import cmd


def test_run_check__counter():
    result = cmd("run", "examples/counter.py", "check")
    assert result.stderr.strip() == ""
    assert result.stdout.strip() == "System check identified no issues (0 silenced)."


def test_run_check__counter_app():
    result = cmd("run", "examples/counter.py:app", "check")
    assert result.stderr.strip() == ""
    assert result.stdout.strip() == "System check identified no issues (0 silenced)."


def test_run_check__counter_module():
    result = cmd(
        "run", "counter:app", "check", cwd="examples", env={"PYTHONPATH": ".."}
    )
    assert result.stderr.strip() == ""
    assert result.stdout.strip() == "System check identified no issues (0 silenced)."


def test_run_check__scale():
    result = cmd("run", "examples/scale.py", "check")
    assert result.stderr.strip() == ""
    assert result.stdout.strip() == "System check identified no issues (0 silenced)."



import urllib.request

from .utils import cmd, nanodjango_process, runserver

TEST_APP = "scale"
TEST_SCRIPT = f"examples/{TEST_APP}.py"
TEST_BIND = "127.0.0.1:8042"


def test_runserver__fbv_with_model():
    cmd("run", TEST_SCRIPT, "makemigrations", TEST_APP)
    cmd("run", TEST_SCRIPT, "migrate")

    with (
        nanodjango_process("run", TEST_SCRIPT, "runserver", TEST_BIND) as handle,
        runserver(handle),
    ):
        response = urllib.request.urlopen(f"http://{TEST_BIND}/", timeout=10)
        assert response.getcode() == 200

        response = urllib.request.urlopen(f"http://{TEST_BIND}/count/", timeout=10)
        assert response.getcode() == 200
        assert "Number of page loads" in response.read().decode("utf-8")



from django.conf import settings
from django.db.migrations.loader import MigrationLoader
from django.db.migrations.writer import MigrationWriter
from django.db.models.base import ModelBase


def patch_modelbase(app_name):
    """
    Because we don't have an app config and don't want to make users set an app label in
    their Meta, pretend they did it anyway
    """
    old_new = ModelBase.__new__

    def new_new(cls, name, bases, attrs, **kwargs):
        module = attrs["__module__"]
        if module == "__main__":
            attrs["__module__"] = app_name
            attr_meta = attrs.get("Meta")
            if attr_meta:
                if not getattr(attr_meta, "app_label", None):
                    attr_meta.app_label = app_name
            else:

                class attr_meta:
                    app_label = app_name

                attrs["Meta"] = attr_meta

        return old_new(cls, name, bases, attrs, **kwargs)

    ModelBase.__new__ = new_new  # type: ignore


def patch_migrations(app_name):
    """
    Migrations needs a root path
    """
    old_basedir = MigrationWriter.basedir.fget
    old_load_disk = MigrationLoader.load_disk

    def new_basedir(self):
        if self.migration.app_label != app_name:
            return old_basedir(self)

        # Ensure
        migrations_dir = settings.BASE_DIR / settings.MIGRATION_MODULES[app_name]
        migrations_dir.mkdir(parents=True, exist_ok=True)
        return str(migrations_dir)

    def new_load_disk(self):
        old_load_disk(self)

    MigrationWriter.basedir = property(new_basedir)  # type: ignore
    MigrationLoader.load_disk = new_load_disk


def patch_db(app_name):
    patch_modelbase(app_name)
    patch_migrations(app_name)



from django.http import HttpResponse


def string_view(fn):
    """
    Wrapper to automatically convert the response from a view function into an
    HttpResponse to support returning a string.
    """

    def django_view(request):
        response = fn(request)
        if isinstance(response, HttpResponse):
            return response
        return HttpResponse(response)

    return django_view
